@@ -1,6 +1,6 @@
 /* continuations.c -- continuations, much stack hackery..
    Copyright (C) 2000 John Harper <john@dcs.warwick.ac.uk>
-   $Id: continuations.c,v 1.28 2000/07/15 11:41:35 john Exp $
+   $Id: continuations.c,v 1.37 2000/09/09 22:38:13 john Exp $
 
    This file is part of librep.
 
@@ -192,8 +192,8 @@ struct rep_barrier_struct {
     rep_thread *head, *tail;
     rep_thread *susp_head, *susp_tail;
     short depth;
-    int closed : 1;
-    int targeted : 1;		/* may contain continuations */
+    u_int closed : 1;
+    u_int targeted : 1;		/* may contain continuations */
 };
 
 /* List of all currently active barriers (on the current stack) */
@@ -212,7 +212,7 @@ struct rep_continuation_struct {
 
     jmp_buf jmpbuf;
     char *stack_copy, *stack_top, *stack_bottom;
-    size_t stack_size;
+    size_t stack_size, real_size;
 
     rep_barrier *barriers;
     rep_barrier *root;
@@ -233,6 +233,8 @@ struct rep_continuation_struct {
 
 #define CF_INVALID	(1 << rep_CELL16_TYPE_BITS)
 
+#define CONTIN_MAX_SLOP 4096
+
 /* the cell16 typecode allocated for continuation objects */
 static int rep_continuation_type;
 
@@ -248,8 +250,8 @@ struct rep_thread_struct {
     repv env, structure;
     int lock;
     struct timeval run_at;
-    rep_bool (*poll)(rep_thread *t, repv arg);
-    repv poll_arg;
+    rep_bool (*poll)(rep_thread *t, void *arg);
+    void *poll_arg;
     repv exit_val;
 };
 
@@ -352,9 +354,8 @@ rep_call_with_barrier (repv (*callback)(repv), repv arg,
 	{
 	    DB (("caught barrier exit throw\n"));
 	    rep_throw_value = rep_CDR (exit_barrier_cell);
-	    if (rep_throw_value == Qnil)
-		rep_throw_value = rep_NULL;
-	    ret = Qnil;
+	    ret = (rep_throw_value == rep_NULL) ? Qnil : rep_NULL;
+	    rep_CDR (exit_barrier_cell) = Qnil;
 	}
 
 	if (rep_throw_value == rep_NULL && b.active != 0)
@@ -432,10 +433,8 @@ common_ancestor (rep_barrier *current, rep_barrier **dest_hist, int dest_depth)
     for (ptr = current; ptr != 0; ptr = ptr->next)
     {
 	int i;
-	DB (("ptr: %p\n", ptr->point));
 	for (i = first_dest; i < dest_depth; i++)
 	{
-	    DB (("dest: %p\n", dest_hist[i]->point));
 	    if (dest_hist[i]->point == ptr->point)
 		return ptr;
 	    else if (SP_NEWER_P (dest_hist[i]->point, ptr->point))
@@ -455,19 +454,44 @@ common_ancestor (rep_barrier *current, rep_barrier **dest_hist, int dest_depth)
 static void
 save_stack (rep_continuation *c)
 {
+    size_t size;
+
     FLUSH_REGISTER_WINDOWS;
 
-#if STACK_DIRECTION < 0
-    c->stack_size = c->stack_bottom - c->stack_top;
+#if defined (__GNUC__) && !defined (BROKEN_ALPHA_GCC)
+    c->stack_top = __builtin_frame_address (0);
 #else
-    c->stack_size = c->stack_top - c->stack_bottom;
+    c->stack_top = (char *) &size;
 #endif
-    c->stack_copy = rep_alloc (c->stack_size);
-    rep_data_after_gc += c->stack_size;
+
 #if STACK_DIRECTION < 0
-    memcpy (c->stack_copy, c->stack_top, c->stack_size);
+    size = c->stack_bottom - c->stack_top;
 #else
-    memcpy (c->stack_copy, c->stack_bottom, c->stack_size);
+    size = c->stack_top - c->stack_bottom;
+#endif
+
+    if (c->stack_copy != 0)
+    {
+	if (c->stack_size < size || (c->stack_size - size) > CONTIN_MAX_SLOP)
+	{
+	    rep_free (c->stack_copy);
+	    rep_data_after_gc -= c->stack_size;
+	    c->stack_copy = 0;
+	}
+    }
+
+    if (c->stack_copy == 0)
+    {
+	c->stack_size = size;
+	c->stack_copy = rep_alloc (size);
+	rep_data_after_gc += size;
+    }
+
+    c->real_size = size;
+#if STACK_DIRECTION < 0
+    memcpy (c->stack_copy, c->stack_top, c->real_size);
+#else
+    memcpy (c->stack_copy, c->stack_bottom, c->real_size);
 #endif
 }
 
@@ -494,9 +518,9 @@ grow_stack_and_invoke (rep_continuation *c, char *water_mark)
        below the current position. */
 
 #if STACK_DIRECTION < 0
-    memcpy (c->stack_top, c->stack_copy, c->stack_size);
+    memcpy (c->stack_top, c->stack_copy, c->real_size);
 #else
-    memcpy (c->stack_bottom, c->stack_copy, c->stack_size);
+    memcpy (c->stack_bottom, c->stack_copy, c->real_size);
 #endif
 
     longjmp (c->jmpbuf, 1);
@@ -572,9 +596,15 @@ DEFUN("primitive-invoke-continuation", Fprimitive_invoke_continuation,
     return rep_NULL;
 }
 
+static repv
+get_cont (repv arg)
+{
+    return Fsymbol_value (Qcontinuation, Qnil);
+}
+
 DEFUN("continuation-callable-p", Fcontinuation_callable_p,
       Scontinuation_callable_p, (repv cont), rep_Subr1) /*
-::doc:continuation-callable-p::
+::doc:rep.lang.interpreter#continuation-callable-p::
 continuation-callable-p CONTINUATION
 
 Returns `t' if the continuation object CONTINUATION from the current
@@ -586,9 +616,11 @@ execution point of the interpreter.
     int depth;
 
     rep_DECLARE1(cont, rep_FUNARGP);
-    cont = rep_FUNARG(cont)->env;
-    rep_DECLARE(1, cont, rep_CAAR (cont) == Qcontinuation);
-    c = rep_CONTIN (rep_CDAR (cont));
+    cont = rep_call_with_closure (cont, get_cont, Qnil);
+    if (cont == rep_NULL)
+	return rep_NULL;
+    rep_DECLARE1(cont, rep_CONTINP);
+    c = rep_CONTIN (cont);
 
     if (c->car & CF_INVALID)
 	return Qnil;
@@ -603,24 +635,11 @@ execution point of the interpreter.
     return anc == 0 ? Qnil : Qt;
 }
 
-/* Return the address further into the stack than any part of the frame
-   of the calling function. */
-static char *
-get_stack_top (rep_continuation *c)
-{
-#if defined (__GNUC__) && !defined (BROKEN_ALPHA_GCC)
-    return __builtin_frame_address (0);
-#else
-    int dummy;
-    return (char *) &dummy;
-#endif
-}
-
 static repv
-primitive_call_cc (repv (*callback)(rep_continuation *, void *), void *data)
+primitive_call_cc (repv (*callback)(rep_continuation *, void *), void *data,
+		   rep_continuation *c)
 {
     struct rep_saved_regexp_data re_data;
-    rep_continuation *c;
     repv ret;
 
     if (root_barrier == 0)
@@ -629,12 +648,17 @@ primitive_call_cc (repv (*callback)(rep_continuation *, void *), void *data)
 	return Fsignal (Qerror, rep_LIST_1 (rep_VAL (&no_root)));
     }
 
-    c = rep_ALLOC_CELL (sizeof (rep_continuation));
-    rep_data_after_gc += sizeof (rep_continuation);
-    c->next = continuations;
-    continuations = c;
-    c->car = rep_continuation_type;
+    if (c == 0)
+    {
+	c = rep_ALLOC_CELL (sizeof (rep_continuation));
+	rep_data_after_gc += sizeof (rep_continuation);
+	c->next = continuations;
+	continuations = c;
+	c->stack_copy = 0;
+    }
 
+    c->car = rep_continuation_type;
+    
     if (setjmp (c->jmpbuf))
     {
 	/* back from call/cc */
@@ -707,12 +731,11 @@ primitive_call_cc (repv (*callback)(rep_continuation *, void *), void *data)
 	c->single_step = rep_single_step_flag;
 	c->lisp_depth = rep_lisp_depth;
 
-	c->stack_top = get_stack_top (c);
 	c->stack_bottom = c->root->point;
 	save_stack (c);
 
-	DB (("call/cc: saved %p; stack_size=%lu (%u)\n",
-	     c, (u_long) c->stack_size, rep_stack_bottom - c->stack_top));
+	DB (("call/cc: saved %p; real_size=%lu (%u)\n",
+	     c, (u_long) c->real_size, rep_stack_bottom - c->stack_top));
 
 	ret = callback (c, data);
 
@@ -734,7 +757,7 @@ inner_call_cc (rep_continuation *c, void *data)
 }
 
 DEFUN("call/cc", Fcall_cc, Scall_cc, (repv fun), rep_Subr1) /*
-::doc:call/cc::
+::doc:rep.lang.interpreter#call/cc::
 call/cc FUNCTION
 
 Invoke FUNCTION with a single parameter, the continuation function of
@@ -744,7 +767,7 @@ control immediately back to the statement following the call to the
 `call/cc' function (even if that stack frame has since been exited).
 ::end:: */
 {
-    return primitive_call_cc (inner_call_cc, (void *) fun);
+    return primitive_call_cc (inner_call_cc, (void *) fun, 0);
 }
 
 
@@ -890,7 +913,7 @@ again:
 	       then invoke the next thread */
 	    active->lock = rep_thread_lock;
 	    thread_save_environ (active);
-	    primitive_call_cc (inner_thread_invoke, active);
+	    primitive_call_cc (inner_thread_invoke, active, active->cont);
 	}
 	else
 	{
@@ -907,8 +930,8 @@ again:
 	if (root_barrier->susp_head == 0)
 	{
 	    root_barrier->active = 0;
-	    if (rep_throw_value)
-		rep_CDR (exit_barrier_cell) = rep_throw_value;
+	    assert (rep_throw_value != exit_barrier_cell);
+	    rep_CDR (exit_barrier_cell) = rep_throw_value;
 	    rep_throw_value = exit_barrier_cell;
 	    DB (("no more threads, throwing to root..\n"));
 	    return;
@@ -973,7 +996,7 @@ new_thread (repv name)
     t->car = rep_thread_type;
     t->name = name;
     t->poll = 0;
-    t->poll_arg = Qnil;
+    t->poll_arg = 0;
     t->exit_val = rep_NULL;
     t->next_alloc = threads;
     threads = t;
@@ -1000,13 +1023,13 @@ make_thread (repv thunk, repv name)
 	thread_save_environ (t);
 	/* this continuation will never get called,
 	   but it simplifies things.. */
-	if (primitive_call_cc (inner_make_thread, x) != -1)
+	if (primitive_call_cc (inner_make_thread, x, 0) != -1)
 	    abort ();
 	root_barrier->active = x;
     }
 
     rep_PUSHGC (gc_thunk, thunk);
-    ret = primitive_call_cc (inner_make_thread, t);
+    ret = primitive_call_cc (inner_make_thread, t, 0);
     rep_POPGC;
     if (ret == -1)
 	return t;
@@ -1018,7 +1041,6 @@ make_thread (repv thunk, repv name)
 	{
 	    t->exit_val = ret;
 	    thread_delete (t);
-	    thread_invoke ();
 	    assert (rep_throw_value == exit_barrier_cell);
 	}
 	else
@@ -1082,7 +1104,7 @@ thread_yield (void)
 
 static void
 thread_suspend (rep_thread *t, u_long msecs,
-		rep_bool (*poll)(rep_thread *t, repv arg), repv poll_arg)
+		rep_bool (*poll)(rep_thread *t, void *arg), void *poll_arg)
 {
     rep_barrier *root = t->cont->root;
     assert (!(t->car & TF_SUSPENDED));
@@ -1212,7 +1234,7 @@ mark_cont (repv obj)
 	if(sd->type == rep_reg_obj)
 	{
 	    int i;
-	    for(i = 0; i < NSUBEXP; i++)
+	    for(i = 0; i < rep_NSUBEXP; i++)
 	    {
 		rep_MARKVAL(sd->matches.obj.startp[i]);
 		rep_MARKVAL(sd->matches.obj.endp[i]);
@@ -1273,7 +1295,6 @@ mark_thread (repv obj)
     rep_MARKVAL (THREAD (obj)->env);
     rep_MARKVAL (THREAD (obj)->structure);
     rep_MARKVAL (THREAD (obj)->name);
-    rep_MARKVAL (THREAD (obj)->poll_arg);
     rep_MARKVAL (THREAD (obj)->exit_val);
 }
 
@@ -1400,7 +1421,7 @@ call_with_outwards (void *data_)
 
 DEFUN("call-with-object", Fcall_with_object,
       Scall_with_object, (repv arg, repv thunk), rep_Subr2) /*
-::call-with-object::
+::doc:rep.lang.interpreter#call-with-object::
 call-with-object ARG THUNK
 
 Call the zero-parameter function THUNK, with object ARG temporarily
@@ -1433,7 +1454,7 @@ unbound. If THUNK is subsequently reentered, ARG will be rebound.
 
 DEFUN("call-with-dynamic-root", Fcall_with_dynamic_root,
       Scall_with_dynamic_root, (repv thunk), rep_Subr1) /*
-::doc:call-with-dynamic-root::
+::doc:rep.lang.interpreter#call-with-dynamic-root::
 call-with-dynamic-root THUNK
 
 Call the zero-parameter function THUNK, as the root of a new execution
@@ -1463,7 +1484,7 @@ call_out (void *data_)
 
 DEFUN("call-with-barrier", Fcall_with_barrier, Scall_with_barrier,
       (repv thunk, repv closed, repv in, repv out), rep_Subr4) /*
-::doc:call-with-barrier::
+::doc:rep.lang.interpreter#call-with-barrier::
 call-with-barrier THUNK CLOSED [IN-THUNK] [OUT-THUNK]
 
 Call THUNK inside a new execution environment. If CLOSED is non-`nil'
@@ -1491,7 +1512,7 @@ The value of this function is the value returned by THUNK.
 }
 
 DEFUN("make-thread", Fmake_thread, Smake_thread, (repv thunk, repv name), rep_Subr2) /*
-::doc:make-thread::
+::doc:rep.threads#make-thread::
 make-thread THUNK [NAME]
 
 Create and return an object representing a new thread of execution. The
@@ -1507,7 +1528,7 @@ parameters.
 }
 
 DEFUN("thread-yield", Fthread_yield, Sthread_yield, (void), rep_Subr0) /*
-::doc:thread-yield::
+::doc:rep.threads#thread-yield::
 thread-yield
 
 Pass control away from the current thread if other threads are waiting
@@ -1522,7 +1543,7 @@ to run.
 }
 
 DEFUN("thread-delete", Fthread_delete, Sthread_delete, (repv th), rep_Subr1) /*
-::doc:thread-delete::
+::doc:rep.threads#thread-delete::
 thread-delete [THREAD]
 
 Mark THREAD (or the current thread), as being deleted. It will not be
@@ -1544,7 +1565,7 @@ thread results forces the containing dynamic root to be closed.
 
 DEFUN("thread-suspend", Fthread_suspend,
       Sthread_suspend, (repv th, repv msecs), rep_Subr2) /*
-::doc:thread-suspend::
+::doc:rep.threads#thread-suspend::
 thread-suspend [THREAD] [MSECS]
 
 Mark THREAD (or the current thread) as being suspended. It will not be
@@ -1557,8 +1578,7 @@ runnable threads, then sleep until the next thread becomes runnable.
     if (th == Qnil)
 	th = Fcurrent_thread (Qnil);
     rep_DECLARE1 (th, THREADP);
-    thread_suspend (THREAD (th), rep_INTP (msecs) ? rep_INT (msecs) : 0,
-		    0, Qnil);
+    thread_suspend (THREAD (th), rep_INTP (msecs) ? rep_INT (msecs) : 0, 0, 0);
     return Qnil;
 #else
     return rep_signal_arg_error (th, 1);
@@ -1567,15 +1587,16 @@ runnable threads, then sleep until the next thread becomes runnable.
 
 #ifdef WITH_CONTINUATIONS
 static rep_bool
-thread_join_poller (rep_thread *t, repv arg)
+thread_join_poller (rep_thread *t, void *arg)
 {
-    return (THREAD (arg)->car & TF_EXITED) ? rep_TRUE : rep_FALSE;
+    rep_thread *th = arg;
+    return (th->car & TF_EXITED) ? rep_TRUE : rep_FALSE;
 }
 #endif
 
 DEFUN("thread-join", Fthread_join,
       Sthread_join, (repv th, repv msecs, repv def), rep_Subr3) /*
-::doc:thread-join::
+::doc:rep.threads#thread-join::
 thread-join THREAD [MSECS] [DEFAULT-VALUE]
 
 Suspend the current thread until THREAD has exited, or MSECS
@@ -1592,7 +1613,7 @@ the last form it evaluated, else return DEFAULT-VALUE.
 	rep_PUSHGC (gc_th, th);
 	thread_suspend (THREAD (self),
 			rep_INTP (msecs) ? rep_INT (msecs) : 0,
-			thread_join_poller, th);
+			thread_join_poller, THREAD (th));
 	rep_POPGC;
 	if ((THREAD (th)->car & TF_EXITED) && THREAD (th)->exit_val)
 	    return THREAD (th)->exit_val;
@@ -1604,7 +1625,7 @@ the last form it evaluated, else return DEFAULT-VALUE.
 }
 
 DEFUN("thread-wake", Fthread_wake, Sthread_wake, (repv th), rep_Subr1) /*
-::doc:thread-wake::
+::doc:rep.threads#thread-wake::
 thread-wake [THREAD]
 
 If THREAD (or the current thread) is currently suspended, mark it as
@@ -1623,7 +1644,7 @@ being runnable once more.
 }
 
 DEFUN("threadp", Fthreadp, Sthreadp, (repv arg), rep_Subr1) /*
-::doc:threadp::
+::doc:rep.threads#threadp::
 threadp ARG
 
 Return `t' if ARG is a thread object.
@@ -1638,7 +1659,7 @@ Return `t' if ARG is a thread object.
 
 DEFUN("thread-suspended-p", Fthread_suspended_p,
       Sthread_suspended_p, (repv th), rep_Subr1) /*
-::doc:thread-suspended-p::
+::doc:rep.threads#thread-suspended-p::
 thread-suspended-p THREAD
 
 Return `t' if THREAD is currently suspended from running.
@@ -1654,7 +1675,7 @@ Return `t' if THREAD is currently suspended from running.
 
 DEFUN("thread-exited-p", Fthread_exited_p,
       Sthread_exited_p, (repv th), rep_Subr1) /*
-::doc:thread-exited-p::
+::doc:rep.threads#thread-exited-p::
 thread-exited-p THREAD
 
 Return `t' if THREAD has exited.
@@ -1670,7 +1691,7 @@ Return `t' if THREAD has exited.
 
 DEFUN("current-thread", Fcurrent_thread,
       Scurrent_thread, (repv depth_), rep_Subr1) /*
-::doc:current-thread::
+::doc:rep.threads#current-thread::
 current-thread [DEPTH]
 
 Return the currently executing thread, or `nil' if threaded execution
@@ -1690,7 +1711,7 @@ is not currently taking place.
 }
 
 DEFUN("all-threads", Fall_threads, Sall_threads, (repv depth_), rep_Subr1) /*
-::doc:all-threads::
+::doc:rep.threads#all-threads::
 all-threads [DEPTH]
 
 Return a list of all threads.
@@ -1717,7 +1738,7 @@ Return a list of all threads.
 }
 
 DEFUN("thread-forbid", Fthread_forbid, Sthread_forbid, (void), rep_Subr0) /*
-::doc:thread-forbid::
+::doc:rep.threads#thread-forbid::
 thread-forbid
 
 Increment the thread preemption lock. When greather than zero all
@@ -1730,7 +1751,7 @@ as this function returns.
 }
 
 DEFUN("thread-permit", Fthread_permit, Sthread_permit, (void), rep_Subr0) /*
-::doc:thread-permit::
+::doc:rep.threads#thread-permit::
 thread-permit
 
 Decrement the thread preemption lock. When greather than zero all
@@ -1743,7 +1764,7 @@ as this function returns.
 }
 
 DEFUN("thread-name", Fthread_name, Sthread_name, (repv th), rep_Subr1) /*
-::doc:thread-name:
+::doc:rep.threads#thread-name:
 thread-name THREAD
 
 Return the name of the thread THREAD.
@@ -1763,6 +1784,8 @@ Return the name of the thread THREAD.
 void
 rep_continuations_init (void)
 {
+    repv tem = rep_push_structure ("rep.lang.interpreter");
+
 #ifdef WITH_CONTINUATIONS
     rep_continuation_type = rep_register_new_type ("continuation",
 						   rep_ptr_cmp,
@@ -1776,7 +1799,7 @@ rep_continuations_init (void)
     exit_barrier_cell = Fcons (Qnil, Qnil);
     rep_mark_static (&exit_barrier_cell);
     rep_INTERN(continuation);
-    rep_ADD_SUBR(Sprimitive_invoke_continuation);
+    rep_ADD_INTERNAL_SUBR(Sprimitive_invoke_continuation);
 #endif
 
     rep_ADD_SUBR(Scall_cc);
@@ -1784,6 +1807,9 @@ rep_continuations_init (void)
     rep_ADD_SUBR(Scall_with_object);
     rep_ADD_SUBR(Scall_with_dynamic_root);
     rep_ADD_SUBR(Scall_with_barrier);
+    rep_pop_structure (tem);
+
+    tem = rep_push_structure ("rep.threads");
     rep_ADD_SUBR(Smake_thread);
     rep_ADD_SUBR(Sthread_yield);
     rep_ADD_SUBR(Sthread_delete);
@@ -1798,4 +1824,5 @@ rep_continuations_init (void)
     rep_ADD_SUBR(Sthread_forbid);
     rep_ADD_SUBR(Sthread_permit);
     rep_ADD_SUBR(Sthread_name);
+    rep_pop_structure (tem);
 }
