@@ -1,6 +1,6 @@
 /* misc.c -- Miscellaneous functions
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: misc.c,v 1.40 2000/05/25 13:49:53 john Exp $
+   $Id: misc.c,v 1.44 2000/09/27 16:36:19 john Exp $
 
    This file is part of Jade.
 
@@ -39,7 +39,7 @@ DEFSYM(process_environment, "process-environment");
 DEFSYM(rep_version, "rep-version");
 DEFSYM(rep_interface_id, "rep-interface-id");
 DEFSYM(rep_build_id, "rep-build-id"); /*
-::doc:operating-system::
+::doc:rep.system#operating-system::
 A symbol defining the type of operating system that Jade is running
 under. Currently this is always the symbol `unix'.
 ::end::
@@ -47,10 +47,10 @@ under. Currently this is always the symbol `unix'.
 A list of all environment variables (as strings "NAME=VALUE") passed
 to the interpreter. Also used to specify the environment of subprocesses.
 ::end::
-::doc:rep-version::
+::doc:rep.system#rep-version::
 A string defining the current version of the REP interpreter.
 ::end::
-::doc:rep-build-id::
+::doc:rep.system#rep-build-id::
 A string describing when, where, and by who the running version of the
 LISP interpreter was built.
 ::end:: */
@@ -63,15 +63,15 @@ DEFSYM(upcase_table, "upcase-table");
 DEFSYM(downcase_table, "downcase-table");
 DEFSYM(flatten_table, "flatten-table");
 /* Some doc strings
-::doc:upcase-table::
+::doc:rep.data#upcase-table::
 256-byte string holding translations to turn each character into its
 upper-case equivalent.
 ::end::
-::doc:downcase-table::
+::doc:rep.data#downcase-table::
 256-byte string holding translations to turn each character into its
 lower-case equivalent.
 ::end::
-::doc:flatten-table::
+::doc:rep.data#flatten-table::
 Translation table to convert newline characters to spaces.
 ::end:: */
 
@@ -90,6 +90,33 @@ stpcpy(register char *dst, register const char *src)
 }
 #endif /* !HAVE_STPCPY */
 
+#ifndef HAVE_STRNCASECMP
+/* Compare no more than N characters of S1 and S2,
+   ignoring case, returning less than, equal to or
+   greater than zero if S1 is lexicographically less
+   than, equal to or greater than S2. (from glibc)  */
+int
+strncasecmp (const char *s1, const char *s2, size_t n)
+{
+  const unsigned char *p1 = (const unsigned char *) s1;
+  const unsigned char *p2 = (const unsigned char *) s2;
+  unsigned char c1, c2;
+
+  if (p1 == p2 || n == 0)
+    return 0;
+
+  do
+    {
+      c1 = tolower (*p1++);
+      c2 = tolower (*p2++);
+      if (c1 == '\0' || c1 != c2)
+        return c1 - c2;
+    } while (--n > 0);
+
+  return c1 - c2;
+}
+#endif
+
 u_char *
 rep_str_dupn(const u_char *old, int len)
 {
@@ -110,7 +137,7 @@ default_beep (void)
 }
 
 DEFUN_INT("beep", Fbeep, Sbeep, (void), rep_Subr0, "") /*
-::doc:beep::
+::doc:rep.system#beep::
 beep
 
 Rings a bell.
@@ -123,7 +150,7 @@ Rings a bell.
 
 DEFUN("complete-string", Fcomplete_string, Scomplete_string,
       (repv existing, repv arg_list, repv fold), rep_Subr3) /*
-::doc:complete-string::
+::doc:rep.data#complete-string::
 complete-string TEMPLATE LIST [FOLD-CASE]
 
 Return a string whose beginning matches the string TEMPLATE, and is unique
@@ -181,7 +208,7 @@ is t, all matching ignores character case.
 }
 
 DEFUN("current-time", Fcurrent_time, Scurrent_time, (void), rep_Subr0) /*
-::doc:current-time::
+::doc:rep.system#current-time::
 current-time
 
 Return a value denoting the current system time. This will be a cons cell
@@ -194,7 +221,7 @@ number of seconds since the start of the day (universal time).
 }
 
 DEFUN("current-utime", Fcurrent_utime, Scurrent_utime, (void), rep_Subr0) /*
-::doc:current-utime::
+::doc:rep.system#current-utime::
 current-utime
 
 Return the current time in microseconds.
@@ -205,7 +232,7 @@ Return the current time in microseconds.
 }
 
 DEFUN("fix-time", Ffix_time, Sfix_time, (repv time), rep_Subr1) /*
-::doc:fix-time::
+::doc:rep.system#fix-time::
 fix-time TIMESTAMP
 
 Ensure that the two parts of TIMESTAMP are mutually consistent. If not
@@ -222,7 +249,7 @@ TIMESTAMP is altered. Returns TIMESTAMP.
 
 DEFUN("current-time-string", Fcurrent_time_string,
       Scurrent_time_string, (repv time, repv format), rep_Subr2) /*
-::doc:current-time-string::
+::doc:rep.system#current-time-string::
 current-time-string [TIME] [FORMAT]
 
 Returns a human-readable string defining the current date and time, or if
@@ -255,7 +282,7 @@ the same conventions as the template to the C library's strftime function.
 }
 
 DEFUN("time-later-p", Ftime_later_p, Stime_later_p, (repv t1, repv t2), rep_Subr2) /*
-::doc:time-later-p::
+::doc:rep.system#time-later-p::
 time-later-p TIME-STAMP1 TIME-STAMP2
 
 Returns t when TIME-STAMP1 refers to a later time than TIME-STAMP2.
@@ -271,7 +298,7 @@ Returns t when TIME-STAMP1 refers to a later time than TIME-STAMP2.
 
 DEFUN("sleep-for", Fsleep_for, Ssleep_for, (repv secs, repv msecs),
       rep_Subr2) /*
-::doc:sleep-for::
+::doc:rep.system#sleep-for::
 sleep-for SECONDS [MILLISECONDS]
 
 Pause for SECONDS (plus the optional MILLISECOND component) length of time.
@@ -284,7 +311,7 @@ Pause for SECONDS (plus the optional MILLISECOND component) length of time.
 
 DEFUN("sit-for", Fsit_for, Ssit_for, (repv secs, repv msecs),
       rep_Subr2) /*
-::doc:sit-for::
+::doc:rep.system#sit-for::
 sit-for [SECONDS] [MILLISECONDS]
 
 Wait for input to arrive and be processed. No more than SECONDS seconds plus
@@ -300,7 +327,7 @@ immediately, using a null timeout.
 }
 
 DEFUN("user-login-name", Fuser_login_name, Suser_login_name, (void), rep_Subr0) /*
-::doc:user-login-name::
+::doc:rep.system#user-login-name::
 user-login-name
 
 Returns the login name of the user (a string).
@@ -310,7 +337,7 @@ Returns the login name of the user (a string).
 }
 
 DEFUN("user-full-name", Fuser_full_name, Suser_full_name, (repv arg), rep_Subr1) /*
-::doc:user-full-name::
+::doc:rep.system#user-full-name::
 user-full-name [REAL-NAME]
 
 Returns the real name of the user (a string). If REAL-NAME is non-nil, it's
@@ -329,7 +356,7 @@ the name to return in subsequent calls.
 
 DEFUN("user-home-directory", Fuser_home_directory,
       Suser_home_directory, (repv user), rep_Subr1) /*
-::doc:user-home-directory::
+::doc:rep.system#user-home-directory::
 user-home-directory [USER]
 
 Return the path to USER's home directory (a string). When USER is undefined
@@ -342,7 +369,7 @@ the directory of the user who executed Jade is found.
 }
 
 DEFUN("system-name", Fsystem_name, Ssystem_name, (void), rep_Subr0) /*
-::doc:system-name::
+::doc:rep.system#system-name::
 system-name
 
 Returns the name of the host which the editor is running on.
@@ -352,7 +379,7 @@ Returns the name of the host which the editor is running on.
 }
 
 DEFUN("message", Fmessage, Smessage, (repv string, repv now), rep_Subr2) /*
-::doc:message::
+::doc:rep.system#message::
 message STRING [DISPLAY-NOW]
 
 Temporarily sets the status display to STRING, this may not happen until the
@@ -385,7 +412,7 @@ next complete redisplay, unless DISPLAY-NOW is non-nil.
 #endif
 
 DEFUN("random", Frandom, Srandom, (repv arg), rep_Subr1) /*
-::doc:random::
+::doc:rep.lang.math#random::
 random [LIMIT]
 
 Produce a pseudo-random number between zero and LIMIT (or the largest positive
@@ -435,7 +462,7 @@ with the current time of day.
 }
 
 DEFUN("translate-string", Ftranslate_string, Stranslate_string, (repv string, repv table), rep_Subr2) /*
-::doc:translate-string:
+::doc:rep.data#translate-string:
 translate-string STRING TRANSLATION-TABLE
 
 Applies the TRANSLATION-TABLE to each character in the string STRING.
@@ -460,11 +487,12 @@ Note that the STRING really is modified, no copy is made!
 	register u_char c = *str;
 	*str++ = (c < tablen) ? rep_STR(table)[c] : c;
     }
+    rep_string_modified (string);
     return(string);
 }
 
 DEFUN("alpha-char-p", Falpha_char_p, Salpha_char_p, (repv ch), rep_Subr1) /*
-::doc:alpha-char-p::
+::doc:rep.data#alpha-char-p::
 alpha-char-p CHAR
 
 Returns t if CHAR is an alphabetic character.
@@ -474,7 +502,7 @@ Returns t if CHAR is an alphabetic character.
 }
 
 DEFUN("upper-case-p", Fupper_case_p, Supper_case_p, (repv ch), rep_Subr1) /*
-::doc:upper-case-p::
+::doc:rep.data#upper-case-p::
 upper-case-p CHAR
 
 Returns t if CHAR is upper case.
@@ -484,7 +512,7 @@ Returns t if CHAR is upper case.
 }
 
 DEFUN("lower-case-p", Flower_case_p, Slower_case_p, (repv ch), rep_Subr1) /*
-::doc:lower-case-p::
+::doc:rep.data#lower-case-p::
 lower-case-p CHAR
 
 Returns t if CHAR is lower case.
@@ -494,7 +522,7 @@ Returns t if CHAR is lower case.
 }
 
 DEFUN("digit-char-p", Fdigit_char_p, Sdigit_char_p, (repv ch), rep_Subr1) /*
-::doc:digit-char-p::
+::doc:rep.data#digit-char-p::
 digit-char-p CHAR
 
 Returns t if CHAR is a digit.
@@ -504,7 +532,7 @@ Returns t if CHAR is a digit.
 }
 
 DEFUN("alphanumericp", Falphanumericp, Salphanumericp, (repv ch), rep_Subr1) /*
-::doc:alphanumericp::
+::doc:rep.data#alphanumericp::
 alphanumericp CHAR
 
 Returns t if CHAR is alpha-numeric.
@@ -514,7 +542,7 @@ Returns t if CHAR is alpha-numeric.
 }
 
 DEFUN("space-char-p", Fspace_char_p, Sspace_char_p, (repv ch), rep_Subr1) /*
-::doc:space-char-p::
+::doc:rep.data#space-char-p::
 space-char-p CHAR
 
 Returns t if CHAR is whitespace.
@@ -524,7 +552,7 @@ Returns t if CHAR is whitespace.
 }
 
 DEFUN("char-upcase", Fchar_upcase, Schar_upcase, (repv ch), rep_Subr1) /*
-::doc:char-upcase::
+::doc:rep.data#char-upcase::
 char-upcase CHAR
 
 Returns the upper-case equivalent of CHAR.
@@ -535,7 +563,7 @@ Returns the upper-case equivalent of CHAR.
 }
 
 DEFUN("char-downcase", Fchar_downcase, Schar_downcase, (repv ch), rep_Subr1) /*
-::doc:char-downcase::
+::doc:rep.data#char-downcase::
 char-downcase CHAR
 
 Returns the lower-case equivalent of CHAR.
@@ -546,7 +574,7 @@ Returns the lower-case equivalent of CHAR.
 }
 
 DEFUN_INT("system", Fsystem, Ssystem, (repv command), rep_Subr1, "sShell command:") /*
-::doc:system::
+::doc:rep.system#system::
 system SHELL-COMMAND
 
 Synchronously execute the shell command string SHELL-COMMAND. Returns the
@@ -563,14 +591,36 @@ values represent that the process was killed by a signal.
     return rep_system (rep_STR (command));
 }
 
+DEFUN("get-command-line-option", Fget_command_line_option,
+      Sget_command_line_option, (repv opt, repv arg), rep_Subr2) /*
+::doc:rep.system#get-command-line-option::
+get-command-line-option OPTION [REQUIRES-ARGUMENT]
+
+Returns t if OPTION was specified on the command line (OPTION is typically
+a word beginning with `--'). If REQUIRES-ARGUMENT is non-nil, this option
+requires a parameter, the value of which is returned. If a parameters isn't
+supplied an error is signalled.
+::end:: */
+{
+    repv param = Qt;
+    rep_DECLARE1(opt, rep_STRINGP);
+    if (rep_get_option (rep_STR(opt), (arg == Qnil) ? 0 : &param))
+	return param;
+    else
+	return Qnil;
+}
+
 void
 rep_misc_init(void)
 {
     int i;
+    repv tem;
 
     if (rep_beep_fun == 0)
 	rep_beep_fun = default_beep;
 
+    tem = rep_push_structure ("rep.system");
+
     rep_INTERN(operating_system);
 #ifdef rep_HAVE_UNIX
     rep_INTERN(unix);
@@ -588,7 +638,6 @@ rep_misc_init(void)
     Fset (Qrep_build_id, rep_VAL(&build_id_string));
 
     rep_ADD_SUBR_INT(Sbeep);
-    rep_ADD_SUBR(Scomplete_string);
     rep_ADD_SUBR(Scurrent_time);
     rep_ADD_SUBR(Scurrent_utime);
     rep_ADD_SUBR(Sfix_time);
@@ -596,13 +645,17 @@ rep_misc_init(void)
     rep_ADD_SUBR(Stime_later_p);
     rep_ADD_SUBR(Ssleep_for);
     rep_ADD_SUBR(Ssit_for);
-
+    rep_ADD_SUBR(Sget_command_line_option);
+    rep_ADD_SUBR_INT(Ssystem);
     rep_ADD_SUBR(Suser_login_name);
     rep_ADD_SUBR(Suser_full_name);
     rep_ADD_SUBR(Suser_home_directory);
     rep_ADD_SUBR(Ssystem_name);
     rep_ADD_SUBR(Smessage);
-    rep_ADD_SUBR(Srandom);
+
+    rep_pop_structure (tem);
+
+    tem = rep_push_structure ("rep.data");
     rep_ADD_SUBR(Stranslate_string);
     rep_ADD_SUBR(Salpha_char_p);
     rep_ADD_SUBR(Supper_case_p);
@@ -612,8 +665,7 @@ rep_misc_init(void)
     rep_ADD_SUBR(Sspace_char_p);
     rep_ADD_SUBR(Schar_upcase);
     rep_ADD_SUBR(Schar_downcase);
-    rep_ADD_SUBR_INT(Ssystem);
-
+    rep_ADD_SUBR(Scomplete_string);
     {
 	repv up = rep_make_string (257);
 	repv down = rep_make_string (257);
@@ -643,4 +695,9 @@ rep_misc_init(void)
 	rep_INTERN(flatten_table);
 	Fset (Qflatten_table, flatten);
     }
+    rep_pop_structure (tem);
+
+    tem = rep_push_structure ("rep.lang.math");
+    rep_ADD_SUBR(Srandom);
+    rep_pop_structure (tem);
 }
