@@ -1,6 +1,6 @@
 /* repdoc.c -- Program to strip doc-strings from C source
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: repdoc.c,v 1.5 2000/03/01 00:27:26 john Exp $
+   $Id: repdoc.c,v 1.6 2000/08/11 19:24:18 john Exp $
 
    This file is part of Jade.
 
@@ -23,6 +23,10 @@
 #include <gdbm.h>
 #include <fcntl.h>
 
+#ifndef GDBM_NOLOCK
+# define GDBM_NOLOCK 0
+#endif
+
 static void
 usage(void)
 {
@@ -79,7 +83,7 @@ main(int ac, char **av)
     av++;
     if(ac < 2)
 	usage();
-    docdbm = gdbm_open(*av++, 0, GDBM_WRCREAT, 0666, 0);
+    docdbm = gdbm_open(*av++, 0, GDBM_WRCREAT | GDBM_NOLOCK, 0666, 0);
     ac--;
     if(docdbm == 0)
     {
