@@ -1,6 +1,6 @@
 /* unix_main.c -- Miscellaneous functions for Unix
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: unix_main.c,v 1.59 2000/07/13 10:44:36 john Exp $
+   $Id: unix_main.c,v 1.63 2000/09/10 10:30:22 john Exp $
 
    This file is part of Jade.
 
@@ -266,6 +266,10 @@ static int input_pending_count;
 void (*rep_register_input_fd_fun)(int fd, void (*callback)(int fd)) = 0;
 void (*rep_deregister_input_fd_fun)(int fd) = 0;
 
+#define MAX_EVENT_LOOP_CALLBACKS 16
+static int next_event_loop_callback;
+static rep_bool (*event_loop_callbacks[MAX_EVENT_LOOP_CALLBACKS])(void);
+
 void
 rep_register_input_fd(int fd, void (*callback)(int fd))
 {
@@ -363,6 +367,27 @@ rep_sig_restart(int sig, rep_bool flag)
 #endif /* !HAVE_SIGINTERRUPT */
 }
 
+void
+rep_add_event_loop_callback (rep_bool (*callback)(void))
+{
+    if (next_event_loop_callback == MAX_EVENT_LOOP_CALLBACKS)
+	abort ();
+    event_loop_callbacks [next_event_loop_callback++] = callback;
+}
+
+rep_bool
+rep_proc_periodically (void)
+{
+    rep_bool ret = rep_FALSE;
+    int i;
+    for (i = 0; i < next_event_loop_callback; i++)
+    {
+	if (event_loop_callbacks[i] ())
+	    ret = rep_TRUE;
+    }
+    return ret;
+}
+
 /* Wait for input for no longer than TIMEOUT-MSECS for input fds defined
    by INPUTS. If input arrived return the number of ready fds, with the
    actual fds defined by the fdset INPUTS. Return zero if the timeout
@@ -442,12 +467,10 @@ wait_for_input(fd_set *inputs, u_long timeout_msecs)
 }
 
 /* Handle the READY fds with pending input (defined by fdset INPUTS).
-   If CALLBACK is non-null, only invoke this function for inputs
-   whose callback function == CALLBACK. Return true if the display
-   might require updating. Returns immediately if a Lisp error has
-   occurred. */
+   Return true if the display might require updating. Returns immediately
+   if a Lisp error has occurred. */
 static rep_bool
-handle_input(fd_set *inputs, int ready, void *callback)
+handle_input(fd_set *inputs, int ready)
 {
     static long idle_period;
     rep_bool refreshp = rep_FALSE;
@@ -463,18 +486,15 @@ handle_input(fd_set *inputs, int ready, void *callback)
 	    if(FD_ISSET(i, inputs))
 	    {
 		ready--;
-		if(!callback || ((void *) input_actions[i]) == callback)
+		if(FD_ISSET(i, &input_pending))
 		{
-		    if(FD_ISSET(i, &input_pending))
-		    {
-			FD_CLR(i, &input_pending);
-			input_pending_count--;
-		    }
-		    if(input_actions[i] != NULL)
-		    {
-			input_actions[i](i);
-			refreshp = rep_TRUE;
-		    }
+		    FD_CLR(i, &input_pending);
+		    input_pending_count--;
+		}
+		if(input_actions[i] != NULL)
+		{
+		    input_actions[i](i);
+		    refreshp = rep_TRUE;
 		}
 	    }
 	}
@@ -511,7 +531,7 @@ rep_event_loop(void)
 
 	memcpy(&copy, &input_fdset, sizeof(copy));
 	ready = wait_for_input(&copy, rep_input_timeout_secs * 1000);
-	refreshp = handle_input(&copy, ready, 0);
+	refreshp = handle_input(&copy, ready);
 
 	/* Check for exceptional conditions. */
 	if(rep_throw_value != rep_NULL)
@@ -551,28 +571,69 @@ rep_sit_for(u_long timeout_msecs)
 }
 
 /* Wait TIMEOUT_MSECS for input, ignoring any input fds that would
-   invoke any callback function except CALLBACK. Return the number of
-   input fds serviced. */
+   invoke any callback function except CALLBACKS. Return Qnil if any
+   input was serviced, Qt if the timeout expired, rep_NULL for an error. */
 repv
-rep_accept_input(u_long timeout_msecs, void *callback)
+rep_accept_input_for_callbacks (u_long timeout_msecs, int ncallbacks,
+				void (**callbacks)(int))
 {
     fd_set copy;
     int ready, i;
     FD_ZERO(&copy);
     for(i = 0; i < FD_SETSIZE; i++)
     {
-	if(FD_ISSET(i, &input_fdset) && ((void *)input_actions[i]) == callback)
-	    FD_SET(i, &copy);
+	if(FD_ISSET(i, &input_fdset))
+	{
+	    int j;
+	    for (j = 0; j < ncallbacks; j++)
+	    {
+		if (input_actions[i] == callbacks[j])
+		{
+		    FD_SET(i, &copy);
+		    break;
+		}
+	    }
+	}
     }
     ready = wait_for_input(&copy, timeout_msecs);
     if(ready > 0 && !rep_INTERRUPTP)
-	handle_input(&copy, ready, callback);
+	handle_input(&copy, ready);
     if(rep_INTERRUPTP)
 	return rep_NULL;
     else
 	return ready > 0 ? Qnil : Qt;
 }
 
+/* Wait TIMEOUT_MSECS for input from the NFDS file descriptors stored in FDS.
+   Return Qnil if any input was serviced, Qt if the timeout expired, rep_NULL
+   for an error. */
+repv
+rep_accept_input_for_fds (u_long timeout_msecs, int nfds, int *fds)
+{
+    fd_set copy;
+    int ready, i;
+    FD_ZERO(&copy);
+    for(i = 0; i < nfds; i++)
+    {
+	if(FD_ISSET(fds[i], &input_fdset))
+	    FD_SET(fds[i], &copy);
+    }
+    ready = wait_for_input(&copy, timeout_msecs);
+    if(ready > 0 && !rep_INTERRUPTP)
+	handle_input(&copy, ready);
+    if(rep_INTERRUPTP)
+	return rep_NULL;
+    else
+	return ready > 0 ? Qnil : Qt;
+}
+
+/* obsolete, for compatibility only */
+repv
+rep_accept_input(u_long timeout_msecs, void (*callback)(int))
+{
+    return rep_accept_input_for_callbacks (timeout_msecs, 1, &callback);
+}
+
 rep_bool
 rep_poll_input(int fd)
 {
@@ -684,7 +745,7 @@ rep_print_allocations(void)
 
 DEFUN("unix-print-allocations", Funix_print_allocations,
       Sunix_print_allocations, (void), rep_Subr0) /*
-::doc:unix-print-allocations::
+::doc:rep.lang.debug#unix-print-allocations::
 unix-print-allocations
 
 Output a list of all allocated memory blocks to standard error.
@@ -881,7 +942,9 @@ rep_sys_os_init(void)
     Fset (Qprocess_environment, env);
 
 #ifdef DEBUG_SYS_ALLOC
-    rep_ADD_SUBR(Sunix_print_allocations);
+    { repv tem = rep_push_structure ("rep.lang.debug");
+      rep_ADD_SUBR(Sunix_print_allocations);
+      rep_pop_structure (tem); }
 #endif
 
     rep_proc_init();
