@@ -1,6 +1,6 @@
 /* unix_processes.c -- Subprocess handling for Unix
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: unix_processes.c,v 1.69 2000/07/10 21:12:44 john Exp $
+   $Id: unix_processes.c,v 1.74 2000/09/14 17:12:45 john Exp $
 
    This file is part of Jade.
 
@@ -229,6 +229,27 @@ proc_notification(void)
     return rep_TRUE;
 }
 
+static inline rep_bool
+notify_queued_p (struct Proc *pr)
+{
+    return pr->pr_NotifyNext != 0;
+}
+
+static void
+notify_1 (struct Proc *pr)
+{
+    if (notify_queued_p (pr))
+    {
+	struct Proc **ptr = &notify_chain;
+	while (*ptr != pr)
+	    ptr = &((*ptr)->pr_NotifyNext);
+	*ptr = pr->pr_NotifyNext;
+	pr->pr_NotifyNext = NULL;
+	if (pr->pr_NotifyFun && pr->pr_NotifyFun != Qnil)
+	    rep_call_lisp1 (pr->pr_NotifyFun, rep_VAL (pr));
+    }
+}
+
 /* Checks if any of my children are zombies, takes appropriate action. */
 static rep_bool
 check_for_zombies(void)
@@ -297,8 +318,8 @@ check_for_zombies(void)
 
 /* Called by the event loop after each event or timeout. Returns true
    if the display should be updated. */
-rep_bool
-rep_proc_periodically(void)
+static rep_bool
+proc_periodically(void)
 {
     rep_bool rc = check_for_zombies();
     if(proc_notification())
@@ -324,7 +345,7 @@ read_from_one_fd(struct Proc *pr, int fd)
 	}
     } while((actual > 0) || (actual < 0 && errno == EINTR));
 
-    if((actual <= 0) && (errno != EWOULDBLOCK) && (errno != EAGAIN))
+    if (actual == 0 || (actual < 0 && errno != EWOULDBLOCK && errno != EAGAIN))
     {
 	/* We assume EOF  */
 
@@ -1017,7 +1038,7 @@ proc_puts(repv stream, void *data, int len, rep_bool is_lisp)
 }
 
 DEFUN("make-process", Fmake_process, Smake_process, (repv stream, repv fun, repv dir, repv prog, repv args), rep_Subr5) /*
-::doc:make-process::
+::doc:rep.io.processes#make-process::
 make-process [OUTPUT-STREAM] [FUN] [DIR] [PROGRAM] [ARGS]
 
 Creates a new process-object, OUTPUT-STREAM is where all output from this
@@ -1071,7 +1092,7 @@ If the DIR parameter is nil it will be inherited from the
 }
 
 DEFUN("start-process", Fstart_process, Sstart_process, (repv arg_list), rep_SubrN) /*
-::doc:start-process::
+::doc:rep.io.processes#start-process::
 start-process [PROCESS] [PROGRAM] [ARGS...]
 
 Starts a process running on process-object PROCESS. The child-process runs
@@ -1145,7 +1166,7 @@ set in the PROCESS prior to calling this function.
 }
 
 DEFUN("call-process", Fcall_process, Scall_process, (repv arg_list), rep_SubrN) /*
-::doc:call-process::
+::doc:rep.io.processes#call-process::
 call-process [PROCESS] [IN-FILE] [PROGRAM] [ARGS...]
 
 Starts a process running on process-object PROCESS. Waits for the child to
@@ -1265,7 +1286,7 @@ do_signal_command(repv proc, int signal, repv signal_group)
 }
 
 DEFUN("interrupt-process", Finterrupt_process, Sinterrupt_process, (repv proc, repv grp), rep_Subr2) /*
-::doc:interrupt-process::
+::doc:rep.io.processes#interrupt-process::
 interrupt-process PROCESS [SIGNAL-GROUP]
 
 Interrupt the asynchronous process PROCESS. If SIGNAL-GROUP is t, interrupt
@@ -1276,7 +1297,7 @@ all child processes of PROCESS (it's process group).
 }
 
 DEFUN("kill-process", Fkill_process, Skill_process, (repv proc, repv grp), rep_Subr2) /*
-::doc:kill-process::
+::doc:rep.io.processes#kill-process::
 kill-process PROCESS [SIGNAL-GROUP]
 
 Kill the asynchronous process PROCESS. If SIGNAL-GROUP is t, kill all
@@ -1287,7 +1308,7 @@ child processes of PROCESS (it's process group).
 }
 
 DEFUN("stop-process", Fstop_process, Sstop_process, (repv proc, repv grp), rep_Subr2) /*
-::doc:stop-process::
+::doc:rep.io.processes#stop-process::
 stop-process PROCESS [SIGNAL-GROUP]
 
 Suspends execution of PROCESS, see `continue-process'. If SIGNAL-GROUP is
@@ -1298,7 +1319,7 @@ non-nil also suspends the processes in the process group of PROCESS.
 }
 
 DEFUN("continue-process", Fcontinue_process, Scontinue_process, (repv proc, repv grp), rep_Subr2) /*
-::doc:continue-process::
+::doc:rep.io.processes#continue-process::
 continue-process PROCESS [SIGNAL-GROUP]
 
 Restarts PROCESS after it has been stopped (via `stop-process'). If
@@ -1326,7 +1347,7 @@ PROCESS.
 
 DEFUN("signal-process", Fsignal_process, Ssignal_process,
       (repv proc, repv sig, repv grp), rep_Subr3) /*
-::doc:signal_process::
+::doc:rep.io.processes#signal_process::
 signal-process PROCESS SIGNAL [SIGNAL-GROUP]
 
 Sends the signal SIGNAL to the process PROCESS. If SIGNAL-GROUP is
@@ -1494,7 +1515,7 @@ the symbol `INT' for the UNIX SIGINT signal.
 }
 
 DEFUN("process-exit-status", Fprocess_exit_status, Sprocess_exit_status, (repv proc), rep_Subr1) /*
-::doc:process-exit-status::
+::doc:rep.io.processes#process-exit-status::
 process-exit-status PROCESS
 
 Returns the unprocessed exit-status of the last process to be run on the
@@ -1512,7 +1533,7 @@ process-object PROCESS. If PROCESS is currently running, return nil.
 }
 
 DEFUN("process-exit-value", Fprocess_exit_value, Sprocess_exit_value, (repv proc), rep_Subr1) /*
-::doc:process-exit-value::
+::doc:rep.io.processes#process-exit-value::
 process-exit-value PROCESS
 
 Returns the return-value of the last process to be run on PROCESS, or nil if:
@@ -1530,7 +1551,7 @@ Returns the return-value of the last process to be run on PROCESS, or nil if:
 }
 
 DEFUN("process-id", Fprocess_id, Sprocess_id, (repv proc), rep_Subr1) /*
-::doc:process-id::
+::doc:rep.io.processes#process-id::
 process-id [PROCESS]
 
 If PROCESS is running or stopped, return the process-identifier associated
@@ -1552,7 +1573,7 @@ If PROCESS is nil, return the process id of the Lisp interpreter.
 }
 
 DEFUN("process-running-p", Fprocess_running_p, Sprocess_running_p, (repv proc), rep_Subr1) /*
-::doc:process-running-p::
+::doc:rep.io.processes#process-running-p::
 process-running-p PROCESS
 
 Return t if PROCESS is running.
@@ -1568,7 +1589,7 @@ Return t if PROCESS is running.
 }
 
 DEFUN("process-stopped-p", Fprocess_stopped_p, Sprocess_stopped_p, (repv proc), rep_Subr1) /*
-::doc:process-stopped-p::
+::doc:rep.io.processes#process-stopped-p::
 process-stopped-p PROCESS
 
 Return t if PROCESS has been stopped.
@@ -1584,7 +1605,7 @@ Return t if PROCESS has been stopped.
 }
 
 DEFUN("process-in-use-p", Fprocess_in_use_p, Sprocess_in_use_p, (repv proc), rep_Subr1) /*
-::doc:process-in-use-p::
+::doc:rep.io.processes#process-in-use-p::
 process-in-use-p PROCESS
 
 Similar to `process-running-p' except that this returns t even when the
@@ -1601,7 +1622,7 @@ process has stopped.
 }
 
 DEFUN("processp", Fprocessp, Sprocessp, (repv arg), rep_Subr1) /*
-::doc:process-p::
+::doc:rep.io.processes#process-p::
 processp ARG
 
 Return t is ARG is a process-object.
@@ -1613,7 +1634,7 @@ Return t is ARG is a process-object.
 }
 
 DEFUN("process-prog", Fprocess_prog, Sprocess_prog, (repv proc), rep_Subr1) /*
-::doc:process-prog::
+::doc:rep.io.processes#process-prog::
 process-prog PROCESS
 
 Return the name of the program in PROCESS.
@@ -1626,7 +1647,7 @@ Return the name of the program in PROCESS.
 }
 
 DEFUN("set-process-prog", Fset_process_prog, Sset_process_prog, (repv proc, repv prog), rep_Subr2) /*
-::doc:set-process-prog::
+::doc:rep.io.processes#set-process-prog::
 set-process-prog PROCESS PROGRAM
 
 Sets the name of the program to run on PROCESS to FILE.
@@ -1639,7 +1660,7 @@ Sets the name of the program to run on PROCESS to FILE.
 }
 
 DEFUN("process-args", Fprocess_args, Sprocess_args, (repv proc), rep_Subr1) /*
-::doc:process-args::
+::doc:rep.io.processes#process-args::
 process-args PROCESS
 
 Return the list of arguments to PROCESS.
@@ -1652,7 +1673,7 @@ Return the list of arguments to PROCESS.
 }
 
 DEFUN("set-process-args", Fset_process_args, Sset_process_args, (repv proc, repv args), rep_Subr2) /*
-::doc:set-process-args::
+::doc:rep.io.processes#set-process-args::
 set-process-args PROCESS ARG-LIST
 
 Set the arguments to PROCESS.
@@ -1666,7 +1687,7 @@ Set the arguments to PROCESS.
 }
 
 DEFUN("process-output-stream", Fprocess_output_stream, Sprocess_output_stream, (repv proc), rep_Subr1) /*
-::doc:process-output-stream::
+::doc:rep.io.processes#process-output-stream::
 process-output-stream PROCESS
 
 Return the stream to which all output from PROCESS is sent.
@@ -1679,7 +1700,7 @@ Return the stream to which all output from PROCESS is sent.
 }
 
 DEFUN("set-process-output-stream", Fset_process_output_stream, Sset_process_output_stream, (repv proc, repv stream), rep_Subr2) /*
-::doc:set-process-output-stream::
+::doc:rep.io.processes#set-process-output-stream::
 set-process-output-stream PROCESS STREAM
 
 Set the output-stream of PROCESS to STREAM. nil means discard all output.
@@ -1691,7 +1712,7 @@ Set the output-stream of PROCESS to STREAM. nil means discard all output.
 }
 
 DEFUN("process-error-stream", Fprocess_error_stream, Sprocess_error_stream, (repv proc), rep_Subr1) /*
-::doc:process-error-stream::
+::doc:rep.io.processes#process-error-stream::
 process-error-stream PROCESS
 
 Return the stream to which all standard-error output from PROCESS is sent.
@@ -1704,7 +1725,7 @@ Return the stream to which all standard-error output from PROCESS is sent.
 }
 
 DEFUN("set-process-error-stream", Fset_process_error_stream, Sset_process_error_stream, (repv proc, repv stream), rep_Subr2) /*
-::doc:set-process-error-stream::
+::doc:rep.io.processes#set-process-error-stream::
 set-process-error-stream PROCESS STREAM
 
 Set the error-stream of PROCESS to STREAM. nil means discard all output.
@@ -1718,7 +1739,7 @@ Note that this currently only works correctly with pipe connections.
 }
 
 DEFUN("process-function", Fprocess_function, Sprocess_function, (repv proc), rep_Subr1) /*
-::doc:process-function::
+::doc:rep.io.processes#process-function::
 process-function PROCESS
 
 Return the function which is called when PROCESS changes state (i.e. it
@@ -1732,7 +1753,7 @@ exits or is stopped).
 }
 
 DEFUN("set-process-function", Fset_process_function, Sset_process_function, (repv proc, repv fn), rep_Subr2) /*
-::doc:set-process-function::
+::doc:rep.io.processes#set-process-function::
 set-process-function PROCESS FUNCTION
 
 Set the function which is called when PROCESS changes state to FUNCTION.
@@ -1744,7 +1765,7 @@ Set the function which is called when PROCESS changes state to FUNCTION.
 }
 
 DEFUN("process-dir", Fprocess_dir, Sprocess_dir, (repv proc), rep_Subr1) /*
-::doc:process-dir::
+::doc:rep.io.processes#process-dir::
 process-dir PROCESS
 
 Return the name of the directory which becomes the working directory of
@@ -1758,7 +1779,7 @@ PROCESS when it is started.
 }
 
 DEFUN("set-process-dir", Fset_process_dir, Sset_process_dir, (repv proc, repv dir), rep_Subr2) /*
-::doc:set-process-dir::
+::doc:rep.io.processes#set-process-dir::
 set-process-dir PROCESS DIR
 
 Set the directory of PROCESS to DIR.
@@ -1781,7 +1802,7 @@ Set the directory of PROCESS to DIR.
 }
 
 DEFUN("process-connection-type", Fprocess_connection_type, Sprocess_connection_type, (repv proc), rep_Subr1) /*
-::doc:process-connection-type::
+::doc:rep.io.processes#process-connection-type::
 process-connection-type PROCESS
 
 Returns a symbol defining the type of stream (i.e. pipe, pty, or
@@ -1795,7 +1816,7 @@ socketpair) used to connect PROCESS with its physical process.
 }
 
 DEFUN("set-process-connection-type", Fset_process_connection_type, Sset_process_connection_type, (repv proc, repv type), rep_Subr2) /*
-::doc:set-process-connection-type::
+::doc:rep.io.processes#set-process-connection-type::
 set-process-connection-type PROCESS TYPE
 
 Define how PROCESS communicates with it's child process, TYPE may be
@@ -1821,7 +1842,7 @@ process error output to be differentiated.
 
 DEFUN("active-processes", Factive_processes, Sactive_processes, (void),
       rep_Subr0) /*
-::doc:active-processes::
+::doc:rep.io.processes#active-processes::
 active-processes
 
 Return a list containing all active process objects.
@@ -1842,9 +1863,20 @@ Return a list containing all active process objects.
     return head;
 }
 
+#define MAX_HANDLERS 16
+static void (*input_handlers[MAX_HANDLERS])(int);
+static int n_input_handlers = 0;
+
+void
+rep_register_process_input_handler (void (*handler)(int))
+{
+    assert (n_input_handlers < MAX_HANDLERS);
+    input_handlers[n_input_handlers++] = handler;
+}
+
 DEFUN("accept-process-output", Faccept_process_output,
       Saccept_process_output, (repv secs, repv msecs), rep_Subr2) /*
-::doc:accept-process-output::
+::doc:rep.io.processes#accept-process-output::
 accept-process-output [SECONDS] [MILLISECONDS]
 
 Wait SECONDS plus MILLISECONDS for output from any asynchronous subprocesses.
@@ -1857,9 +1889,10 @@ Note that output includes notification of process termination.
     /* Only wait for output if nothing already waiting. */
     if(!got_sigchld && !notify_chain)
     {
-	result = rep_accept_input((rep_INTP(secs) ? rep_INT(secs) * 1000 : 0)
-				  + (rep_INTP(msecs) ? rep_INT(msecs) : 0),
-				  (void *)&read_from_process);
+	result = (rep_accept_input_for_callbacks
+		  ((rep_INTP(secs) ? rep_INT(secs) * 1000 : 0)
+		   + (rep_INTP(msecs) ? rep_INT(msecs) : 0),
+		   n_input_handlers, input_handlers));
     }
     if(got_sigchld || notify_chain)
     {
@@ -1869,6 +1902,48 @@ Note that output includes notification of process termination.
     return result;
 }
 
+DEFUN("accept-process-output-1", Faccept_process_output_1,
+      Saccept_process_output_1,
+      (repv process, repv secs, repv msecs), rep_Subr3) /*
+::doc:rep.io.processes#accept-process-output-1::
+accept-process-output-1 PROCESS [SECONDS] [MILLISECONDS]
+
+Wait SECONDS plus MILLISECONDS for output from the asynchronous
+subprocess PROCESS. If any arrives, process it, then return nil.
+Otherwise return t.
+
+Note that output includes notification of process termination.
+::end:: */
+{
+    repv result = Qt;
+    rep_DECLARE1 (process, PROCESSP);
+
+    /* Only wait for output if nothing already waiting. */
+    if (got_sigchld)
+	check_for_zombies ();
+
+    if (!notify_queued_p (VPROC (process)))
+    {
+	int fds[2];
+	fds[0] = VPROC (process)->pr_Stdout;
+	fds[1] = VPROC (process)->pr_Stderr;
+	result = (rep_accept_input_for_fds
+		  ((rep_INTP(secs) ? rep_INT(secs) * 1000 : 0)
+		   + (rep_INTP(msecs) ? rep_INT(msecs) : 0), 2, fds));
+    }
+
+    if (got_sigchld)
+	check_for_zombies ();
+
+    if (notify_queued_p (VPROC (process)))
+    {
+	notify_1 (VPROC (process));
+	result = Qt;
+    }
+
+    return result;
+}
+
 /* Don't use libc system (), since it blocks signals. */
 repv
 rep_system (char *command)
@@ -1934,6 +2009,8 @@ rep_system (char *command)
 void
 rep_proc_init(void)
 {
+    repv tem;
+
     /* Setup SIGCHLD stuff.  */
     sigemptyset(&chld_sigset);
     sigaddset(&chld_sigset, SIGCHLD);
@@ -1952,6 +2029,8 @@ rep_proc_init(void)
     rep_INTERN(pipe);
     rep_INTERN(pty);
     rep_INTERN(socketpair);
+
+    tem = rep_push_structure ("rep.io.processes");
     rep_ADD_SUBR(Smake_process);
     rep_ADD_SUBR(Sstart_process);
     rep_ADD_SUBR(Scall_process);
@@ -1983,12 +2062,16 @@ rep_proc_init(void)
     rep_ADD_SUBR(Sset_process_connection_type);
     rep_ADD_SUBR(Sactive_processes);
     rep_ADD_SUBR(Saccept_process_output);
+    rep_ADD_SUBR(Saccept_process_output_1);
+    rep_pop_structure (tem);
 
     process_type = rep_register_new_type ("subprocess", rep_ptr_cmp,
 					  proc_prin, proc_prin,
 					  proc_sweep, proc_mark,
 					  mark_active_processes,
 					  0, 0, proc_putc, proc_puts, 0, 0);
+    rep_register_process_input_handler (read_from_process);
+    rep_add_event_loop_callback (proc_periodically);
 }
 
 void
